// Copyright 2017, OpenCensus Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package readme

import (
	"context"

	"go.opencensus.io/trace"
)

func traceExamples() {
	ctx := context.Background()

	// START startend
	ctx, span := trace.StartSpan(ctx, "cache.Get")
	defer span.End()

	// Do work to get from cache.
	// END startend
}
